@@ -159,6 +159,10 @@ export_completed_with_success (DialogData *data)
 	builder = _gtk_builder_new_from_file ("flicker-export-completed.ui", "flicker");
 	dialog = _gtk_builder_get_widget (builder, "completed_messagedialog");
 	g_object_set_data_full (G_OBJECT (dialog), "builder", builder, g_object_unref);
+	g_signal_connect (dialog,
+			  "delete-event",
+			  G_CALLBACK (gtk_true),
+			  NULL);
 	g_signal_connect (dialog,
 			  "response",
 			  G_CALLBACK (completed_messagedialog_response_cb),
@@ -535,6 +539,7 @@ dlg_export_to_flickr (FlickrServer *server,
 	gth_file_list_set_thumb_size (GTH_FILE_LIST (data->list_view), 112);
 	gth_file_view_set_spacing (GTH_FILE_VIEW (gth_file_list_get_view (GTH_FILE_LIST (data->list_view))), 0);
 	gth_file_list_enable_thumbs (GTH_FILE_LIST (data->list_view), TRUE);
+	gth_file_list_set_ignore_hidden (GTH_FILE_LIST (data->list_view), TRUE);
 	gth_file_list_set_caption (GTH_FILE_LIST (data->list_view), "none");
 	gth_file_list_set_sort_func (GTH_FILE_LIST (data->list_view), gth_main_get_sort_type ("file::name")->cmp_func, FALSE);
 	gtk_widget_show (data->list_view);
@@ -552,7 +557,7 @@ dlg_export_to_flickr (FlickrServer *server,
 
 	g_signal_connect (data->dialog,
 			  "delete-event",
-			  G_CALLBACK (gtk_widget_hide_on_delete),
+			  G_CALLBACK (gtk_true),
 			  NULL);
 	g_signal_connect (data->dialog,
 			  "response",
