@@ -16,8 +16,7 @@
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 /* This was based on a file from Tracker */
@@ -394,7 +393,7 @@ extract_metadata (MetadataExtractor *extractor,
         if ((extractor->video_height >= 0) && (extractor->video_width >= 0))
                 add_metadata (info,
                 	      "general::dimensions",
-                	      g_strdup_printf ("%d × %d", (guint) extractor->video_width, (guint) extractor->video_height),
+                	      g_strdup_printf (_("%d × %d"), (guint) extractor->video_width, (guint) extractor->video_height),
                 	      NULL);
 
         if ((extractor->video_fps_n >= 0) && (extractor->video_fps_d >= 0))
